#!/usr/bin/env python

import sys
import os

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

# import .

from jgtutils import (
    jgtconstants as constants,
    jgtcommon as jgtcommon,
    jgtwslhelper as wsl,
)

import argparse
import JGTPDSP as pds
import JGTCDS as cds
import JGTCDSSvc as svc
import JGTADS as ads
from JGTChartConfig import JGTChartConfig

import  JGTADSRequest as RQ

import pandas as pd


def parse_args():
    parser = argparse.ArgumentParser(description="Process command parameters.")
    # jgtfxcommon.add_main_arguments(parser)
    jgtcommon.add_instrument_timeframe_arguments(parser)
    jgtcommon.add_date_arguments(parser)
    jgtcommon.add_tlid_range_argument(parser)
    jgtcommon.add_max_bars_arguments(parser)
    # jgtcommon.add_output_argument(parser)
    # jgtfxcommon.add_quiet_argument(parser)
    jgtcommon.add_verbose_argument(parser)
    jgtcommon.add_ads_argument(parser)
    jgtcommon.add_use_full_argument(parser)
    jgtcommon.add_use_fresh_argument(parser)
    
    jgtcommon.add_ids_mfi_argument(parser)
    jgtcommon.add_ids_gator_oscillator_argument(parser)
    jgtcommon.add_ids_balligator_argument(parser)
    jgtcommon.add_ids_talligator_argument(parser)
    jgtcommon.add_ids_fractal_largest_period_argument(parser)
    jgtcommon.add_viewpath_argument(parser)
    
    # jgtcommon.add_cds_argument(parser)
    args = parser.parse_args()
    return args


def main():
    cc = JGTChartConfig()
    args = parse_args()

    gator_oscillator_flag = (
        args.gator_oscillator_flag if args.gator_oscillator_flag else False
    )
    mfi_flag = True #mfi_flag = args.mfi_flag if args.mfi_flag else False
    balligator_flag = args.balligator_flag if args.balligator_flag else False
    talligator_flag = args.talligator_flag if args.talligator_flag else False
    balligator_period_jaws = args.balligator_period_jaws
    talligator_period_jaws = args.talligator_period_jaws
    largest_fractal_period = args.largest_fractal_period


    viewpath=args.viewpath
    
    instrument = args.instrument
    timeframe = args.timeframe
    quotescount = args.quotescount
    #print("#@STCIssue : quotes_count is not passed , it should be !")
    cc.nb_bar_on_chart = quotescount
    
    verbose_level = args.verbose
    quiet = False
    if verbose_level == 0:
        quiet = True
    
    full = False
    fresh = False
    if args.fresh:
        fresh=True
    
    if args.full:
        full = True

    date_from = None
    date_to = None
    tlid_range = None
    if args.tlidrange:
        # @STCGoal Get range prices from cache or request new
        tlid_range = args.tlidrange
        print("#FUTURE Support for tlid range")
        tmpcmd = wsl._mkbash_cmd_string_jgtfxcli_range(
            instrument, timeframe, tlid_range,verbose_level=verbose_level,
            use_full=full
        )
        print(tmpcmd)
        print("#-----------Stay tune -------- Quitting for now")
        return

    if args.datefrom:
        date_from = args.datefrom.replace("/", ".")
    if args.dateto:
        date_to = args.dateto.replace("/", ".")

    show_ads = False
    if args.ads:
        show_ads = True

    # process_cds=args.cds
    process_cds = True
    # output=False
    # compress=False

    # print("Verbose level : " + str(verbose_level))
    if process_cds:
        if verbose_level > 0:print("Processing CDS")
        output = True
    # if args.compress:
    #    compress = args.compress
    #     output = True # in case
    # if args.output:
    #     output = True

    if verbose_level > 1:
        if date_from:
            print("Date from : " + str(date_from))
        if date_to:
            print("Date to : " + str(date_to))

    try:

        print_quiet(quiet, "Getting for : " + instrument + "_" + timeframe)
        instruments = instrument.split(",")
        timeframes = timeframe.split(",")

        for instrument in instruments:
            for timeframe in timeframes:
                createCDS_for_main(
                    instrument,
                    timeframe,
                    quiet=quiet,
                    verbose_level=verbose_level,
                    tlid_range=tlid_range,
                    show_ads=show_ads,
                    cc=cc,
                    use_full=full,
                    use_fresh=fresh,
                    gator_oscillator_flag=gator_oscillator_flag,
                    mfi_flag=mfi_flag,
                    balligator_flag=balligator_flag,
                    balligator_period_jaws=balligator_period_jaws,
                    largest_fractal_period=largest_fractal_period,
                    talligator_flag=talligator_flag,
                    talligator_period_jaws=talligator_period_jaws,
                    viewpath=viewpath,
                    quotescount=quotescount
                )
                # else:
                #     p = pds.getPH(instrument, timeframe, quotes_count, date_from, date_to, False, quiet)
                #     if verbose_level > 0:
                #         print(p)

    except Exception as e:
        jgtcommon.print_exception(e)

    # try:
    #    jgtpy.off()
    # except Exception as e:
    #    jgtfxcommon.print_exception(e)


# if __name__ == "__main__":
#     main()

# print("")
# #input("Done! Press enter key to exit\n")


def createCDS_for_main(
    instrument,
    timeframe,
    quiet,
    verbose_level=0,
    tlid_range=None,
    show_ads=False,
    cc: JGTChartConfig = None,
    use_full=False,
    use_fresh=True,
    gator_oscillator_flag=False,
    mfi_flag=True,
    balligator_flag=False,
    balligator_period_jaws=89,
    largest_fractal_period=89,
    talligator_flag=False,
    talligator_period_jaws=377,
    viewpath=False,
    quotescount=300,
):
    # implementation goes here
    col2remove = constants.columns_to_remove
    config = jgtcommon.readconfig()
    if "columns_to_remove" in config:  # read it from config otherwise
        col2remove = config["columns_to_remove"]
    quietting = True
    if verbose_level > 1:
        quietting = False
        
    try:
        #cdspath, cdf = cds.createFromPDSFileToCDSFile(
        cdspath, cdf = cds.createFromPDSFileToCDSFile(
            instrument, 
            timeframe, 
            quiet=quietting,
            #cc=cc,
            use_full=use_full,
            use_fresh=use_fresh,
            columns_to_remove=col2remove,
            keep_bid_ask=True,
            gator_oscillator_flag=gator_oscillator_flag,
            mfi_flag=mfi_flag,
            balligator_flag=balligator_flag,
            balligator_period_jaws=balligator_period_jaws,
            largest_fractal_period=largest_fractal_period,
            talligator_flag=talligator_flag,
            talligator_period_jaws=talligator_period_jaws,
            viewpath=viewpath,
            quotescount=quotescount
        )  # @STCIssue: This is not supporting -c NB_BARS_TO_PROCESS, should it ?
        if cdspath is not None and cdf is None and viewpath:
            return #we printed it already.
        #@STCGoal GENERATE THE ZONE from the FRESH CDF
        #print("Zone updating...")
        fpath,zone_data = svc.zone_update_from_cdf(instrument,timeframe,cdf,quiet=quietting)
        print("    ",instrument," ",timeframe," zone: ",zone_data["zcol"].values[0])
        
        print_quiet(quiet, cdspath)
        print_quiet(quiet, cdf)
    except Exception as e:
        print("Failed to create CDS for : " + instrument + "_" + timeframe)
        print("jgtcli::Exception in cds.createFromPDSFileToCDSFile(: " + str(e))
        
    try:
        if (
            show_ads
        ):  # (data,instrument,timeframe,nb_bar_on_chart = 375,show=True,plot_ao_peaks=False)
            
            rq=RQ.JGTADSRequest()
            rq.instrument=instrument
            rq.timeframe=timeframe
            rq.show=True
            rq.cc=cc
            _chart,_arr,_data = ads.plot_v2(rq)
            #ads.plot_from_cds_df(
            #    cdf, instrument, timeframe, show=True, plot_ao_peaks=True, cc=cc
            #)
    except Exception as e:
        print("ADS Failed to plot CDS for : " + instrument + "_" + timeframe)
        print("jgtcli::ADS::Exception in  _chart,_arr,_data = ads.plot_v2(rq): " + str(e))

def print_quiet(quiet, content):
    if not quiet:
        print(content)


if __name__ == "__main__":
    main()



# %% Imports
import os

from dotenv import load_dotenv

# Load environment variables from .env file
load_dotenv()

import plotly.graph_objects as go
from plotly.subplots import make_subplots
import panel as pn

import matplotlib.pyplot as plt

plt.rcParams["figure.max_open_warning"] = 100

import pandas as pd
from jgtpy import JGTADS as ads, adshelper as ah, JGTPDSP as pds, JGTCDS as cds
from jgtpy import JGTChartConfig

import tlid

# @STCGoal An online Platform to View charting

# %% Plot all charts in one operation
cc = JGTChartConfig.JGTChartConfig()
cc.saucer_marker_size = 14
cc.ac_signals_marker_size = 14
cc.fig_ratio_x = 31
cc.fig_ratio_y = 16
cc.nb_bar_on_chart = 300
cc.plot_style = "yahoo"


# %% The Experiment

instruments = "GBP/USD,AUD/USD,EUR/USD,SPX500,XAU/USD,USD/CAD,USD/JPY,EUR/CAD,AUD/CAD"
instruments = "GBP/USD,AUD/USD,EUR/USD,SPX500,XAU/USD,USD/CAD,USD/JPY,EUR/CAD,AUD/CAD"
instruments = (
    "AUS200,GBP/USD,AUD/USD,EUR/USD,SPX500,XAU/USD,USD/CAD,USD/JPY,EUR/CAD,AUD/CAD"
)


# Read the INSTRUMENTS variable from the environment or use the default value
instruments = os.getenv("I", instruments)
print(instruments)


defaulttimeframes = "M1,W1,D1,H8,H4,H1,m15,m5"
# Read the TIMEFRAMES variable from the environment or use the default value
timeframes = os.getenv("T", defaulttimeframes).split(",")
print(timeframes)


ldirdef = "./charts"
o = os.getenv("LDIR", ldirdef)
html_outdir_root = o

show_tabs = False
show_chart = False

perspectives = {}
ptabs = pn.Tabs(width=2550, height=1150)

for i in instruments.split(","):
    try:
        print(f"-------------{i}-------------------")

        figures = {}
        success = False
        for t in timeframes:
            print(i, t)
            f, ax, _ = ads.plot(i, t, show=show_chart, cc=cc)
            f.title = t
            figures[t] = f
            povfn = i.replace("/", "-") + "_" + t
            fnout = o + "/" + povfn + ".png"
            fnoutcsv = o + "/" + povfn + ".cds.csv"
            f.savefig(fnout)
            _.to_csv(fnoutcsv)

        # Create a tabbed layout
        # @STCIssue Ratio is different between this and CC
        tabs = pn.Tabs(width=2500, height=1100)

        # Add a tab for each timeframe
        for t in timeframes:
            tabs.append((t, figures[t]))

        if show_tabs:
            tabs.show()

        tabs.title = i

        html_fname = i.replace("/", "-") + ".html"
        html_output_filepath = f"{html_outdir_root}/pto-pers-" + html_fname

        # perspectives[i] = tabs
        tabs.save(html_output_filepath, embed=True)

        perspectives[i] = tabs

        ptabs.append((i, tabs))
    except:
        print("An error occurred while processing:", i)
        pass

# %% Full perspective with more than 1 instrument

full_html_output_filepath = f"{html_outdir_root}/pto-full.html"
print(full_html_output_filepath)

ptabs.save(full_html_output_filepath, embed=True)

# %%



import json
class JGTBaseRequest:
    def __init__(self, quiet=True, verbose_level=0,viewpath=False):
        self.quiet = quiet
        self.verbose_level = verbose_level
        self.viewpath = viewpath
   
    # create a new JGTBaseRequest object from args (argparse)
    @staticmethod
    def from_args(args):
        return JGTBaseRequest(
            quiet=args.quiet,
            verbose_level=args.verbose,
            viewpath=args.viewpath
        )
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2)
    




import sys
import os
import json


sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

from JGTIDSRequest import JGTIDSRequest

class JGTCDSRequest(JGTIDSRequest):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs) 
        
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2)
    # def __str__(self) -> str:
    #     return super().__str__() + f"dummy_cds_flag: {self.dummy_cds_flag}\n" 




import sys
import os

import json

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

from JGTPDSRequest import JGTPDSRequest

from jgtutils.jgtconstants import (TJAW_PERIODS, TTEETH_PERIODS)

class JGTIDSRequest(JGTPDSRequest):
    """
    Represents a JGT IDS (Indicator Data Structure) request.

    Args:
        aof_flag (bool): Flag indicating whether to include aof_flag (AO Peak V1 (Would be replaced by the scipy.signals peak)) data.
        balligator_flag (bool): Flag indicating whether to include Big Alligator data.
        mfi_flag (bool): Flag indicating whether to include mfi_flag (Money Flow Index) data.
        gator_oscillator_flag (bool): Flag indicating whether to include Gator Oscillator data.
        balligator_period_jaws (int): Period for the jaws line of the Alligator indicator.
        balligator_period_teeth (int): Period for the teeth line of the Alligator indicator.
        balligator_period_lips (int): Period for the lips line of the Alligator indicator.
        balligator_shift_jaws (int): Shift for the jaws line of the Alligator indicator.
        balligator_shift_teeth (int): Shift for the teeth line of the Alligator indicator.
        balligator_shift_lips (int): Shift for the lips line of the Alligator indicator.
        largest_fractal_period (int): Period for identifying the largest fractal.
        peak_distance (int): Distance between peaks.
        peak_width (int): Width of the peaks.
        peak_divider_min_height (int): Minimum height of the peak divider.
        rounding_decimal_min (int): Minimum decimal for rounding.
        disable_ao_peaks_v1 (bool): Flag indicating whether to disable AO Peak V1.
        include_ao_color (bool): Flag indicating whether to include AO color data.
        include_ac_color (bool): Flag indicating whether to include AC color data.
        addAlligatorOffsetInFutur (bool): Flag indicating whether to add Alligator offset in the future.
        talligator_flag (bool): Flag indicating whether to include Talligator data.
        talligator_period_jaws (int): Period for the jaws line of the Talligator indicator.
        talligator_period_teeth (int): Period for the teeth line of the Talligator indicator.
        talligator_period_lips (int): Period for the lips line of the Talligator indicator.
        talligator_shift_jaws (int): Shift for the jaws line of the Talligator indicator.
        talligator_shift_teeth (int): Shift for the teeth line of the Talligator indicator.
        talligator_shift_lips (int): Shift for the lips line of the Talligator indicator.
        
        
    """
    def __init__(self, 
                 aof_flag=False, 
                 balligator_flag=False, 
                 mfi_flag=True, 
                 gator_oscillator_flag=False,
                 balligator_period_jaws=89, 
                 balligator_period_teeth=55, 
                 balligator_period_lips=34,
                 balligator_shift_jaws=55, 
                 balligator_shift_teeth=34, 
                 balligator_shift_lips=21,
                 largest_fractal_period=89,
                 peak_distance=13,
                 peak_width=8,
                 peak_divider_min_height = 3,
                 rounding_decimal_min=11, 
                 disable_ao_peaks_v1=True,
                 include_ao_color=False,
                 include_ac_color=False,
                 addAlligatorOffsetInFutur=False,
                 talligator_flag=False, 
                 talligator_period_jaws=377, 
                 talligator_period_teeth=233, 
                 talligator_period_lips=144,
                 talligator_shift_jaws=233, 
                 talligator_shift_teeth=144, 
                 talligator_shift_lips=89,
                 *args, 
                 **kwargs):
        #super().__init__(None, None, None)
        super().__init__(*args, **kwargs)
        self.include_ao_color = include_ao_color
        self.include_ac_color = include_ac_color
        self.disable_ao_peaks_v1 = disable_ao_peaks_v1
        self.aof_flag = aof_flag
        self.balligator_flag = balligator_flag
        self.mfi_flag = mfi_flag
        self.gator_oscillator_flag = gator_oscillator_flag
        self.balligator_period_jaws = balligator_period_jaws
        self.balligator_period_teeth = balligator_period_teeth
        self.balligator_period_lips = balligator_period_lips
        self.balligator_shift_jaws = balligator_shift_jaws
        self.balligator_shift_teeth = balligator_shift_teeth
        self.balligator_shift_lips = balligator_shift_lips
        self.largest_fractal_period = largest_fractal_period
        self.rounding_decimal_min = rounding_decimal_min
        self.peak_distance = peak_distance
        self.peak_width = peak_width
        self.peak_divider_min_height = peak_divider_min_height
        self.addAlligatorOffsetInFutur = addAlligatorOffsetInFutur
        self.talligator_flag=talligator_flag
        self.talligator_period_jaws = talligator_period_jaws
        self.talligator_period_teeth = talligator_period_teeth
        self.talligator_period_lips = talligator_period_lips
        self.talligator_shift_jaws = talligator_shift_jaws
        self.talligator_shift_teeth = talligator_shift_teeth
        self.talligator_shift_lips = talligator_shift_lips

        #Migrated logics
        self.balligator_period_jaws = self.balligator_period_jaws 
        #if self.balligator_flag else 0 #balligator_period_jaws will be 0 if it is not used
        self.talligator_period_jaws = self.talligator_period_jaws 
        #if self.talligator_flag else 0 #talligator_period_jaws will be 0 if it is not used
        
        if self.talligator_flag:
            self.talligator_fix_quotescount()
        else:
            if self.balligator_flag:
                self.balligator_fix_quotescount()
    
    # create a new JGTIDSRequest object from args (argparse)
    @staticmethod
    def from_args(args):
        return JGTIDSRequest(
            instrument=args.instrument,
            timeframe=args.timeframe,
            quotescount=args.quotescount,
            viewpath=args.viewpath,
            use_fresh=args.fresh,
            use_full=args.full,
            gator_oscillator_flag=args.gator_oscillator_flag,
            mfi_flag=args.mfi_flag,
            balligator_flag=args.balligator_flag,
            balligator_period_jaws=args.balligator_period_jaws,
            largest_fractal_period=args.largest_fractal_period,
            talligator_flag=args.talligator_flag,
            talligator_period_jaws=args.talligator_period_jaws,
            verbose_level=args.verbose
        )   
    def to_json(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=2) 
    
    def _get_talligator_required_additional_quotescount(self):            
        TJAW_REQUIRED_CALC_BARS = TJAW_PERIODS+TTEETH_PERIODS #@STCIssue We should use : self.talligator_period_jaws ,... instead of TJAW_PERIODS, TTEETH_PERIODS
        return TJAW_REQUIRED_CALC_BARS

    def talligator_fix_quotescount(self,nb_bars_by_default=300):
        if self.use_full:
            return
        
        if self.talligator_flag:
            if self.quotescount==-1:
                self.quotescount = nb_bars_by_default
            TALLIGATOR_REQ_QUOTECOUNT=self._get_talligator_required_additional_quotescount()
            #print("self.quotescount:",self.quotescount)
            self.quotescount = TALLIGATOR_REQ_QUOTECOUNT + self.quotescount
            #print("self.quotescount:",self.quotescount)
            # if self.quotescount < TALLIGATOR_REQ_QUOTECOUNT + 300:
            #     self.quotescount = TALLIGATOR_REQ_QUOTECOUNT
    
    def _get_balligator_required_additional_quotescount(self):            
        BJAW_REQUIRED_CALC_BARS = self.balligator_period_jaws+self.balligator_shift_jaws 
        return BJAW_REQUIRED_CALC_BARS

    def balligator_fix_quotescount(self,nb_bars_by_default=300):
        if self.use_full:
            return
        
        if self.balligator_flag and not self.talligator_flag:
            if self.quotescount==-1:
                self.quotescount = nb_bars_by_default
            BALLIGATOR_REQ_QUOTECOUNT=self._get_balligator_required_additional_quotescount()
            #print("self.quotescount:",self.quotescount)
            self.quotescount = BALLIGATOR_REQ_QUOTECOUNT + self.quotescount
            


import requests
import json
import os

import sys
import os
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

import jgt2312

class JGTPDSProxyClient:
    def __init__(self, base_url):
        self.base_url = base_url

    def getPH(self, instrument:str, timeframe:str):
        data = {"instrument": instrument, "timeframe": timeframe}
        response = requests.post(f"{self.base_url}/getPH", json=data)
        return response.text

    def getPH_from_filestore(
        self, instrument, timeframe, quiet=True, compressed=False, with_index=True
    ):
        data = {
            "instrument": instrument,
            "timeframe": timeframe,
            "quiet": quiet,
            "compressed": compressed,
            "with_index": with_index,
        }
        response = requests.post(f"{self.base_url}/getPH_from_filestore", json=data)
        return response.json()

    def h(
        self, instrument, timeframe
    ):
        data = {
            "instrument": instrument,
            "timeframe": timeframe
        }
        response = requests.post(f"{self.base_url}/h", json=data)
        return response.json()

    def cli(
        self,
        instrument,
        timeframe,
        datefrom=None,
        dateto=None,
        quote_count=335,
        verbose=0,
    ):
        data = {
            "instrument": instrument,
            "timeframe": timeframe,
            "datefrom": datefrom,
            "dateto": dateto,
            "quote_count": quote_count,
            "verbose": verbose,
        }
        response = requests.post(f"{self.base_url}/cli", json=data)
        return response.json()

    def cli2(
        self,
        instrument:str,
        timeframe:str,
    ):
        data = {
            "instrument": instrument,
            "timeframe": timeframe
        }
        response = requests.post(f"{self.base_url}/cli", json=data)
        print(response)
        return response.json()
    
    def run_init(self):        
        response = requests.get(f"{self.base_url}/init")
        return response.json()

    def fetch_mk_fn(self, instrument:str, timeframe:str, ext:str):
        params = {"instrument": instrument, "timeframe": timeframe, "ext": ext}
        response = requests.get(f"{self.base_url}/mk_fn", params=params)
        return response.json()

    def fetch_get_instrument_properties(self, instrument:str):
        data = {"instrument": instrument.replace("-", "/")}
        response = requests.post(f"{self.base_url}/iprop", json=data)
        return response.json()



import pandas as pd
from jgtutils.jgtconstants import MFI,MFI_SQUAT,MFI_GREEN,MFI_FADE,MFI_FAKE,MFI_SIGNAL,MFI_VAL,MFI_SQUAT_STR,MFI_FAKE_STR,MFI_FADE_STR,MFI_GREEN_STR,MFI_SQUAT_ID,MFI_FAKE_ID,MFI_FADE_ID,MFI_GREEN_ID

print("WARNING::jgtpy.mfihelper MIGHT Migrate to jgtml for now")

def mfi_str_to_id(mfi_str:str):
    if mfi_str == MFI_SQUAT_STR:
        return MFI_SQUAT_ID
    elif mfi_str == MFI_FAKE_STR:
        return MFI_FAKE_ID
    elif mfi_str == MFI_FADE_STR:
        return MFI_FADE_ID
    elif mfi_str == MFI_GREEN_STR:
        return MFI_GREEN_ID
    else:
        return 0

def mfi_signal_to_str(mfi_signal:int):
    if mfi_signal == MFI_SQUAT:
        return MFI_SQUAT_STR
    elif mfi_signal == MFI_FAKE:
        return MFI_FAKE_STR
    elif mfi_signal == MFI_FADE:
        return MFI_FADE_STR
    elif mfi_signal == MFI_GREEN:
        return MFI_GREEN_STR
    else:
        return "0"

def mfi_id_to_str(mfi_id:int):
    """
    Convert the MFI ID to its string representation.
    
    Parameters:
    mfi_id (int): The MFI ID to convert to its string representation.
    
    Returns:
    str: The string representation of the MFI ID.
    
    Possible values:
    - MFI_SQUAT_ID: MFI_SQUAT_STR
    - MFI_FAKE_ID: MFI_FAKE_STR
    - MFI_FADE_ID: MFI_FADE_STR
    - MFI_GREEN_ID: MFI_GREEN_STR
    - 0: "0"
    
    """
    if mfi_id == MFI_SQUAT_ID:
        return MFI_SQUAT_STR
    elif mfi_id == MFI_FAKE_ID:
        return MFI_FAKE_STR
    elif mfi_id == MFI_FADE_ID:
        return MFI_FADE_STR
    elif mfi_id == MFI_GREEN_ID:
        return MFI_GREEN_STR
    else:
        return "0"

def get_mfi_features_column_list_by_timeframe(t:str):
    """
    Get the list of columns that are MFI features for the given timeframe and its related timeframes.
    
    Parameters:
    t (str): The timeframe to get the MFI features columns for.
    
    Returns:
    list: The list of columns that are MFI features for the given timeframe and its related timeframes
    """
    mfi_str_selected_columns = [MFI_VAL+'_M1',MFI_VAL+'_W1']
    
    if t=='H4' or t=='H8' or t=='H6' or t=='H1' or t=='m15' or t=='m5':
      mfi_str_selected_columns.append(MFI_VAL+'_D1')
      
    if t=='H1' or t=='m15' or t=='m5':
        mfi_str_selected_columns.append(MFI_VAL+'_H4')
        
    if t=='m15' or t=='m5':
        mfi_str_selected_columns.append(MFI_VAL+'_H1')
    
    if t=='m5':
        mfi_str_selected_columns.append(MFI_VAL+'_m15')
        
    mfi_str_selected_columns.append(MFI_VAL)
    return mfi_str_selected_columns


def column_mfi_str_in_dataframe_to_id(df:pd.DataFrame,t:str):
    """
    Convert the MFI string columns in the dataframe to their corresponding MFI ID.
    
    Parameters:
    df (pd.DataFrame): The dataframe to convert the MFI string columns to their corresponding MFI ID.
    t (str): The timeframe to get the MFI features columns for.
    
    Returns:
    pd.DataFrame: The dataframe with the MFI string columns converted to their corresponding MFI ID.
    """
    mfi_str_selected_columns=get_mfi_features_column_list_by_timeframe(t)
    for col_name in mfi_str_selected_columns:
        #check if the column exists in the dataframe
        if col_name not in df.columns:
            continue
        df[col_name] = df[col_name].apply(lambda x: int(mfi_str_to_id(x)))
    return df

def column_mfi_str_back_to_str_in_dataframe(df:pd.DataFrame,t:str):
    mfi_str_selected_columns=get_mfi_features_column_list_by_timeframe(t)
    for col_name in mfi_str_selected_columns:
        #check if the column exists in the dataframe
        if col_name not in df.columns:
            continue
        df[col_name] = df[col_name].apply(lambda x: mfi_id_to_str(x)).copy()
    return df


import warnings

# Ignore FutureWarning
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys
import os

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))


import pandas as pd
import datetime
from JGTPDHelper import (
    jgtpd_drop_col_by_name,
    __cleanse_ao_peak_v1_secondary_columns,
    _pds_cleanse_original_columns,
    pds_cleanse_extra_columns,
)

from jgtapy import Indicators

from scipy.signal import find_peaks
from aohelper import add_ao_price_peaks_v2

from JGTChartConfig import JGTChartConfig
from JGTIDSRequest import JGTIDSRequest

# %%
# @title Vars
_dtformat = "%m.%d.%Y %H:%M:%S"

# %%
# @title INDICATOR's Data Frame Columns naming


from jgtutils.jgtconstants import *

from jgtutils.jgtconstants import (
    JAW,
    BJAW,
    TEETH,
    BTEETH,
    LIPS,
    BLIPS,
    OPEN,
    HIGH,
    LOW,
    CLOSE,
    BAR_HEIGHT,
    FH,
    FH3,
    FH5,
    FL,
    FL3,
    FL5,
    FH8,
    FL8,
    FDB,
    FDBB,
    FDBS,
    ACB,
    ACS,
    SB,
    SS,
    AO,
    AC,
    PRICE_PEAK_ABOVE,
    PRICE_PEAK_BELLOW,
    AO_PEAK_ABOVE,
    AO_PEAK_BELLOW,
    FH13,
    FH21,
    FH34,
    FH55,
    FH89,
    FL13,
    FL21,
    FL34,
    FL55,
    FL89,
    MFI
)


# %%
# @title Range shift add col drop na
# --@STCGoal PDS Utils
def _jgtpd_col_add_range_shifting_dropnas(
    dfsrc,
    ctxcolname=indicator_AO_awesomeOscillator_column_name,
    colprefix="pao",
    endrange=10,
):
    return _jgtpd_dropnas_on_any_rows(
        _jgtpd_col_add_range_shifting(dfsrc, ctxcolname, colprefix, endrange)
    )


# @title BACKWARD Range shift col
def _jgtpd_col_add_range_shifting(
    dfsrc,
    ctxcolname=indicator_AO_awesomeOscillator_column_name,
    colprefix="pao",
    endrange=10,
):
    """Add a BACKWARD range of shifted values
    for a column with a prefixed numbered.

    Args:
         dfsrc (DataFrame source)
         ctxcolname (column name from)
         colprefix (new columns prefix)
         endrange (the end of the range from 0)

    Returns:
      DataFrame with new columns
    """
    for i in range(endrange):
        dfsrc[colprefix + str(i)] = dfsrc[ctxcolname].shift(i)
    return dfsrc


# %%
# --@STCGoal IDS Indicators and related / CDS


def normalize_columns(df: pd.DataFrame, columns: list, in_place=True) -> pd.DataFrame:
    if in_place:
        df_normalized = df
        for column in columns:
            df.loc[:, column] = df[column] / df[column].abs().max()
        return df
    else:
        df_normalized = df.copy()
        for column in columns:
            df_normalized.loc[:, column] = df[column] / df[column].abs().max()
    return df_normalized


columns_to_normalize = ["ao", "ac"]


def ids_add_indicatorsV2(
    dfsrc,
    dropnavalue=True,
    quiet=True,
    cleanupOriginalColumn=True,
    cc: JGTChartConfig = None,
    ids_request: JGTIDSRequest = None,
):
    """
    #@STCIssue Backward Compatibility ??
    #@STCGoal What is the ultimate goal ?
      #@STCGoal Well Migrated to the new version of the function
        #@STCIssue   JGTChartConfig/JGTIDSRequest Unclear about Creation of IDS and What Goes in the Chart (ADS)
          #@a ADSRequest vs IDSRequest/CDSRequest
    """
    return None


def ids_add_indicators(
    dfsrc,
    dropnavalue=True,
    quiet=True,
    cleanupOriginalColumn=True,
    useLEGACY=True,
    cc: JGTChartConfig = None,
    bypass_index_reset=False,
    rq: JGTIDSRequest = None,
):
    """
    Adds technical indicators to a given DataFrame.

    Args:
    dfsrc (pandas.DataFrame): The DataFrame to which the indicators will be added.
    dropnavalue (bool, optional): Whether to drop rows with NaN values. Defaults to True.
    quiet (bool, optional): Whether to suppress console output. Defaults to False.
    cleanupOriginalColumn (bool, optional): Whether to clean up the original column. Defaults to True.
    useLEGACY (bool, optional): Whether to use the legacy version of the function. Defaults to True.
    cc (JGTChartConfig, optional): The JGTChartConfig object. Defaults to None.
    bypass_index_reset (bool, optional): Whether to bypass resetting the index. Defaults to False.
    rq (JGTIDSRequest, optional): The JGTIDSRequest object. Defaults to None.

    Returns:
    pandas.DataFrame: The DataFrame with the added indicators.
    """

    if cc is None:
        cc = JGTChartConfig()
    if rq is None:
        rq = JGTIDSRequest()

    dfresult = None

    #Migrated from Signature to Request Object
    enablegator_oscillator_flag=rq.gator_oscillator_flag
    enablemfi_flag=rq.mfi_flag
    big_alligator=rq.balligator_flag
    balligator_period_jaws=rq.balligator_period_jaws
    largest_fractal_period=rq.largest_fractal_period

    if (
        not useLEGACY
    ):  # Because jgtapy has to be upgraded with new column name, we wont use it until our next release
        dfresult = Indicators.jgt_create_ids_indicators_as_dataframe(
            dfsrc,
            enablegator_oscillator_flag,
            enablemfi_flag,
            cleanupOriginalColumn,
            quiet,
        )
    else:
        dfresult = ids_add_indicators_LEGACY(
            dfsrc=dfsrc,
            enablegator_oscillator_flag=enablegator_oscillator_flag,
            enablemfi_flag=enablemfi_flag,
            dropnavalue=dropnavalue,
            quiet=quiet,
            min_nb_bar_on_chart=cc.min_bar_on_chart,
            bypass_index_reset=bypass_index_reset,
            big_alligator=big_alligator,
            ids_request=rq,
            balligator_period_jaws=balligator_period_jaws,
            largest_fractal_period=largest_fractal_period,
        )

    return round_columns(dfresult, rq.rounding_decimal_min)


def round_columns(df, rounding_decimal_min=10):
    for col in df.columns:
        if df[col].dtype == "float64" and df[col].apply(lambda x: x % 1 != 0).any():
            df[col] = df[col].round(decimals=rounding_decimal_min)
            df[col] = df[col].apply(lambda x: 0 if "e" in str(x) else x)
    return df


def round_columns_v2(df, rounding_decimal_min=10):
    df = df.copy()
    float_cols = df.select_dtypes(include=["float64"]).columns
    for col in float_cols:
        if df[col].apply(lambda x: x % 1 != 0).any():
            df[col] = df[col].round(decimals=rounding_decimal_min)
            df[col] = df[col].apply(lambda x: 0 if "e" in str(x) else x)
    return df


def ids_add_indicators_LEGACY(
    dfsrc,
    enablegator_oscillator_flag=False,
    enablemfi_flag=True,
    dropnavalue=True,
    quiet=True,
    addAlligatorOffsetInFutur=False,
    big_alligator=False,
    balligator_period_jaws=89,
    largest_fractal_period=89,
    min_nb_bar_on_chart=300,
    bypass_index_reset=False,
    ids_request: JGTIDSRequest = None,
):
    """
    Adds various technical indicators to the input DataFrame. Is the same as in the jgtapy.legacy module.

    Args:
    dfsrc (pandas.DataFrame): The input DataFrame.
    enablegator_oscillator_flag (bool, optional): Whether to enable the Gator Oscillator indicator. Defaults to False.
    enablemfi_flag (bool, optional): Whether to enable the Money Flow Index indicator. Defaults to False.
    dropnavalue (bool, optional): Whether to drop rows with NaN values. Defaults to True.
    quiet (bool, optional): Whether to suppress print statements. Defaults to False.
    addAlligatorOffsetInFutur (bool, optional): (NOT IMPLEMENTED) Whether to add the Alligator offset in the future. Defaults to True.
    big_alligator_jaws_period (int, optional): (NOT IMPLEMENTED) The period of the Alligator jaws. Defaults to 89.
    largest_fractal_period (int, optional): (NOT IMPLEMENTED) The largest fractal period. Defaults to 89.
    min_nb_bar_on_chart (int, optional): The minimum number of bars on the chart. Defaults to 300.
    bypass_index_reset (bool, optional): Whether to bypass resetting the index. Defaults to False.
    ids_request (JGTIDSRequest, optional): The JGTIDSRequest object. Defaults to None.

    Returns:
    pandas.DataFrame: The input DataFrame with added technical indicators.
    """

    if ids_request is None:
        ids_request = JGTIDSRequest()

    ldfsrc = len(dfsrc)
    # TODO
    df_nb_bars = min_nb_bar_on_chart  # @STCIssue: We have charts with as few as 80 bars and some indicators wont work. We need to find a way to make it work with less bars.
    df_nb_bars = (
        ldfsrc  # @a workaround for now - adequate limits is the amount of bars we have.
    )
    minimal_bars_with_indicators = 25
    FIX_TOO_SHORT_DATAFRAME__PROTO = False

    if FIX_TOO_SHORT_DATAFRAME__PROTO:
        dfsrc = fix_too_short_df(dfsrc)

    if not quiet:
        print("Adding indicators...")
    i = Indicators(dfsrc)
    # print("IDS::debug len(dfsrc)" + str(ldfsrc))

    try:
        i.ao_ac_oscillator(
            column_name_ao=AO,
            column_name_ac=AC,
        )
    except:
        print("ao_ac_oscillator failed")

    try:
        i.alligator(
            period_jaws=13,
            period_teeth=8,
            period_lips=5,
            shift_jaws=8,
            shift_teeth=5,
            shift_lips=3,
            column_name_jaws=JAW,
            column_name_teeth=TEETH,
            column_name_lips=LIPS,
        )
    except:
        print("alligator failed")

    # Assign numbers to variables with prefix 'balligator_'
    balligator_period_jaws = ids_request.balligator_period_jaws
    balligator_period_teeth = ids_request.balligator_period_teeth
    balligator_period_lips = ids_request.balligator_period_lips
    balligator_shift_jaws = ids_request.balligator_shift_jaws
    balligator_shift_teeth = ids_request.balligator_shift_teeth
    balligator_shift_lips = ids_request.balligator_shift_lips

    bAlligator_required_bar_offset = (
        minimal_bars_with_indicators + balligator_period_jaws + balligator_shift_jaws
    )

    if ldfsrc >= bAlligator_required_bar_offset and big_alligator:

        try:
            i.alligator(
                period_jaws=balligator_period_jaws,
                period_teeth=balligator_period_teeth,
                period_lips=balligator_period_lips,
                shift_jaws=balligator_shift_jaws,
                shift_teeth=balligator_shift_teeth,
                shift_lips=balligator_shift_lips,
                column_name_jaws=BJAW,
                column_name_teeth=BTEETH,
                column_name_lips=BLIPS,
            )
        except:
            print("big_alligator failed")
    else:
        if not quiet:
            print("Skipping degree larger big Alligator")

    # Creating Fractal Indicators for degrees 2,3,5,8,13,21,34,55,89

    try:
        i.fractals(
            column_name_high=FH,
            column_name_low=FL,
        )
    except:
        print("fractals failed")
    try:
        i.fractals3(
            column_name_high=FH3,
            column_name_low=FL3,
        )
    except:
        print("fractals 3 failed")
    try:
        i.fractals5(
            column_name_high=FH5,
            column_name_low=FL5,
        )
    except:
        print("fractals 5 failed")
    try:
        i.fractals8(
            column_name_high=FH8,
            column_name_low=FL8,
        )
    except:
        print("fractals 8 failed")
    try:
        i.fractals13(
            column_name_high=FH13,
            column_name_low=FL13,
        )
    except:
        print("fractals 13 failed")
    if (
        df_nb_bars >= minimal_bars_with_indicators + 21
    ):  # @a The amount of bars the chart has to have to calculate the indicator is the minimal number of bars we want with indicators + the period of the indicator
        try:
            i.fractals21(
                column_name_high=FH21,
                column_name_low=FL21,
            )
        except:
            print("fractals 21 failed")

    if df_nb_bars >= minimal_bars_with_indicators + 34:
        try:
            i.fractals34(
                column_name_high=FH34,
                column_name_low=FL34,
            )
        except:
            print("fractals 34 failed")
    else:
        if not quiet:
            print("Skipping Fractal 34")

    if df_nb_bars >= minimal_bars_with_indicators + 55:
        try:
            i.fractals55(
                column_name_high=FH55,
                column_name_low=FL55,
            )
        except:
            print("fractals 55 failed")
    else:
        if not quiet:
            print("Skipping Fractal 55")

    largest_fractal_bar_required_offset = (
        minimal_bars_with_indicators + largest_fractal_period
    )
    if df_nb_bars >= largest_fractal_bar_required_offset:
        try:
            i.fractals89(
                column_name_high=FH89,
                column_name_low=FL89,
            )
        except:
            print("fractals 89 failed")
    else:
        if not quiet:
            print("Skipping Fractal 89")

    if enablegator_oscillator_flag:

        try:
            i.gator(
                period_jaws=13,
                period_teeth=8,
                period_lips=5,
                shift_jaws=8,
                shift_teeth=5,
                shift_lips=3,
                column_name_val1=GL,
                column_name_val2=GH,
            )
        except:
            print("gator failed")

    if enablemfi_flag:
        try:
            i.bw_mfi(column_name=MFI)
        except:
            print("bw_mfi failed")

    if addAlligatorOffsetInFutur:
        try:
            _add_alligator_tmpcol_offset_in_futur(i)
        except:
            print("addAlligatorOffsetInFutur failed")

    dfresult = i.df

    if dropnavalue:
        dfresult = dfresult.dropna()
    try:
        dfresult.set_index("Date", inplace=True)
    except TypeError:
        pass

    normalize = True
    if normalize:
        dfresult = normalize_columns(dfresult, columns_to_normalize)

    if not quiet:
        print("done adding indicators :)")

    # if addAlligatorOffsetInFutur:
    #   _offset_alligator_tmpcol_in_futur(dfresult)
    return dfresult


def fix_too_short_df(dfsrc):
    """
    Fixes a DataFrame that is too short for calculating indicators.

    Args:
    dfsrc (pandas.DataFrame): The input DataFrame.

    Returns:
    pandas.DataFrame: The fixed DataFrame.
    """
    # print("IDS::debug len(dfsrc)" + str(len(dfsrc)))
    if len(dfsrc) < 110:

        # Get the 'Date' value from the first row
        start_date = dfsrc.iloc[0]["Date"]

        # Create a date range starting from the start_date
        date_range = pd.date_range(start=start_date, periods=110)
        # Create a DataFrame with the first row replicated 110 times
        dfsrc_first_row = pd.DataFrame([dfsrc.iloc[0]] * 110)

        # Add the date range to the DataFrame
        dfsrc_first_row["Date"] = date_range

        # Concatenate the two DataFrames
        dfsrc = pd.concat([dfsrc_first_row, dfsrc], ignore_index=True)
    return dfsrc


def _offset_alligator_tmpcol_in_futur(dfsrc):
    dfsrc["jaws_tmp2"] = dfsrc["jaws_tmp"].shift(8)
    dfsrc["teeth_tmp2"] = dfsrc["teeth_tmp"].shift(5)
    dfsrc["lips_tmp2"] = dfsrc["lips_tmp"].shift(3)

    return dfsrc


def _add_alligator_tmpcol_offset_in_futur(i):

    i.smma(8, "jaws_tmp", "Median")
    i.smma(5, "teeth_tmp", "Median")
    i.smma(3, "lips_tmp", "Median")
    # df=i.df
    # Define the logic for adding alligator offset in the future
    # ...
    # indicator_currentDegree_alligator_jaw_column_name
    # from jgtapy.utils import  calculate_smma
    # df_j = calculate_smma(df['Median'], 8, 'jaws_tmp', median_col)
    # df_t = calculate_smma(df_median, period_teeth, column_name_teeth, median_col)
    # df_l = calculate_smma(df_median, period_lips, column_name_lips, median_col)

    # # Shift SMMAs
    # df_j[column_name_jaws] = df_j[column_name_jaws].shift(shift_jaws)
    # df_t[column_name_teeth] = df_t[column_name_teeth].shift(shift_teeth)
    # df_l[column_name_lips] = df_l[column_name_lips].shift(shift_lips)

    return i


# %%
# @title Pandas JGT Utilities


def _jgtpd_dropnas_on_any_rows(dfsrc):
    return dfsrc.dropna(axis="rows")


def _jgtpd_drop_cols_from_to_by_name(dfsrc, firstcolname, lastcolname, _axis=1):
    try:
        return dfsrc.drop(dfsrc.loc[:, firstcolname:lastcolname].columns, axis=_axis)
    except:
        return dfsrc


def _jgtpd_col_drop_range(dfsrc, colprefix="pao", endrange=10):
    firstcolname = colprefix + str(1)
    lastcolname = colprefix + str(endrange)
    return _jgtpd_drop_cols_from_to_by_name(dfsrc, firstcolname, lastcolname, 1)


def _ids_add_fdb_intermediaries_columns(dfsrc):
    """
    Adds intermediate columns to the given DataFrame for the purpose of
    identifying fractal bullish and bearish bars.

    Args:
    - dfsrc: A pandas DataFrame containing OHLC data.

    Returns:
    - A pandas DataFrame with additional columns for intermediate calculations.
    """

    # Bullish
    dfsrc["HighisBellowLips"] = dfsrc.lips > dfsrc.High

    dfsrc["LowIsLower"] = dfsrc.Low < dfsrc.Low.shift()

    dfsrc["ClosedAboveMedian"] = dfsrc.Close > dfsrc.Median

    # Bearish FDBS

    dfsrc["LowisAboveLips"] = dfsrc.lips < dfsrc.Low

    dfsrc["HighIsHigher"] = dfsrc.High > dfsrc.High.shift()

    dfsrc["ClosedBellowMedian"] = dfsrc.Close < dfsrc.Median
    return dfsrc


def _ids_clear_fdb_intermediaries_columns(dfsrc, quiet=False):
    """
    This function drops FDB columns from a pandas dataframe that are not needed for further processing.signalCode_fractalDivergentBar_column_name

    Args:
    - dfsrc: pandas dataframe
    - quiet: boolean, default False. If True, suppresses the print statement when KeyError is caught.

    Returns:
    - dfsrc: pandas dataframe with the specified columns dropped.
    """
    try:
        dfsrc = _jgtpd_drop_cols_from_to_by_name(
            dfsrc, "HighisBellowLips", "ClosedBellowMedian"
        )
    except KeyError:
        if not quiet:
            print("Might cleared already")
        pass
    return dfsrc




def _ids_add_fdb_column_logics_v2(dfsrc, quiet=False):
    dfsrc["fdbb"] = (
        (dfsrc.lips > dfsrc.High)
        & (dfsrc.Low < dfsrc.Low.shift())
        & (dfsrc.Close > dfsrc.Median)
    ).astype(int)
    dfsrc["fdbs"] = (
        (dfsrc.lips < dfsrc.Low)
        & (dfsrc.High > dfsrc.High.shift())
        & (dfsrc.Close < dfsrc.Median)
    ).astype(int)
    dfsrc["fdb"] = dfsrc.apply(
        lambda row: 1 if row["fdbb"] > 0 else -1 if row["fdbs"] > 0 else 0, axis=1
    )
    return dfsrc


def _ids_add_fdb_column_logics(dfsrc, _dropIntermediariesColumns=True, quiet=False):
    """
    Adds FDB (Fractal Divergence Buy) and FDS (Fractal Divergence Sell) columns to the input dataframe based on certain conditions.

    Args:
    - dfsrc: pandas dataframe, input dataframe to which FDB and FDS columns need to be added.
    - _dropIntermediariesColumns: bool, default True. If True, intermediary columns added during the calculation will be dropped.
    - quiet: bool, default False. If True, suppresses the print statements.

    Returns:
    - dfsrc: pandas dataframe, dataframe with FDB and FDS columns added.
    """
    dfsrc = _ids_add_fdb_intermediaries_columns(dfsrc)
    for i, row in dfsrc.iterrows():

        ClosedAboveMedian = dfsrc.at[i, "ClosedAboveMedian"]
        LowIsLower = dfsrc.at[i, "LowIsLower"]
        HighisBellowLips = dfsrc.at[i, "HighisBellowLips"]
        ClosedBellowMedian = dfsrc.at[i, "ClosedBellowMedian"]
        HighIsHigher = dfsrc.at[i, "HighIsHigher"]
        LowisAboveLips = dfsrc.at[i, "LowisAboveLips"]

        # default values
        dfsrc.at[i, FDBB] = 0
        dfsrc.at[i, FDBS] = 0
        dfsrc.at[i, FDB] = 0

        ##################################################
        #########   FDBB
        isFDB = 0
        isFDBCode = 0
        high = 0
        low = 0
        if HighisBellowLips and LowIsLower and ClosedAboveMedian:
            isFDB = 1
            isFDBCode = 1
            high = dfsrc.at[i, HIGH]
            low = dfsrc.at[i, LOW]
        dfsrc.at[i, "fdbbhigh"] = high
        dfsrc.at[i, "fdbblow"] = low

        dfsrc.at[i, FDBB] = isFDB
        dfsrc.at[i, FDB] = isFDBCode  # So we have All
        isAfdbb = isFDB
        ##################################################
        #########   FDBS

        if not isAfdbb:
            isFDB = 0
            isFDBCode = 0
            high = 0
            low = 0
            if LowisAboveLips and HighIsHigher and ClosedBellowMedian:
                isFDB = 1
                isFDBCode = -1
                high = dfsrc.at[i, HIGH]
                low = dfsrc.at[i, LOW]

            dfsrc.at[i, "fdbshigh"] = high
            dfsrc.at[i, "fdbslow"] = low

            dfsrc.at[i, FDBS] = isFDB
            dfsrc.at[i, FDB] = isFDBCode
    if _dropIntermediariesColumns:
        dfsrc = _ids_clear_fdb_intermediaries_columns(dfsrc, quiet=quiet)
    return dfsrc


# @title aof_flag function pto (AO Fractals)


# @title Range shift col
def _add_ao_fractal_peak_v1(
    dfsrc,
    ctxcolname=indicator_AO_awesomeOscillator_column_name,
    poscolprefix="pao",
    negcolprefix="nao",
    endrange=10,
    quiet=False,
    cc: JGTChartConfig = None,
):
    """Make the AO Fractal Peak

    Args:
         dfsrc (DataFrame source)
         ctxcolname (column name from)
         poscolprefix (prefix positive (futur) ao sec col )
         negcolprefix (prefix negative (past) ao sec col )
         endrange (total range from 0 (zero being the current in the output))
         quiet (quiet mode)
         cc (JGTChartConfig, optional): The JGTChartConfig object. Defaults to None.

    Returns:
      DataFrame with new AO Peak columns
        aof (21,13,0,-13,-21) Fractal value of the peak
        aofvalue (AO Value)
        aofhighao (AO value on Bullish Peak)
        aoflowao (AO value of Bearish Peak)
        aofhigh  (Price High of that peak)
        aoflow   (Price Low of that Peak)

    """

    half = endrange / 2
    l_df = len(dfsrc)
    for o in range(endrange):
        i = int(o - half)

        _cn = poscolprefix + str(i)
        if i < 0:
            _cn = negcolprefix + str(i).replace("-", "")
        dfsrc[_cn] = dfsrc[ctxcolname].shift(i)

    # This filters out some noisy peaks that are too low or too high
    df_standardDeviation = dfsrc[ctxcolname].std()
    df_max = dfsrc[ctxcolname].max()
    df_min = dfsrc[ctxcolname].min()
    df_filterOutaof_flagThreshold_ABOVE = (df_standardDeviation + df_max) / 2
    df_filterOutaof_flagThreshold_BELLOW = ((df_standardDeviation * -1) + df_min) / 2

    if not quiet:
        print(
            "filterout (std) Above(max)/(min)Bellow:  ("
            + str(df_standardDeviation)
            + ")  "
            + str(df_filterOutaof_flagThreshold_ABOVE)
            + "("
            + str(df_max)
            + ") / ("
            + str(df_min)
            + ") "
            + str(df_filterOutaof_flagThreshold_BELLOW)
        )

    # Counting the peaks
    countUpPeak = 0
    countDownPeak = 0
    countDiscarted = 0
    for i, row in dfsrc.iterrows():
        baraof_flag = 0
        dt = i
        curHigh = dfsrc.at[i, "High"]
        curLow = dfsrc.at[i, "Low"]
        cur = dfsrc.at[i, indicator_AO_awesomeOscillator_column_name]
        n9 = dfsrc.at[i, "n9"]
        n13 = dfsrc.at[i, "n13"]
        n12 = dfsrc.at[i, "n12"]
        n11 = dfsrc.at[i, "n11"]
        n10 = dfsrc.at[i, "n10"]
        n8 = dfsrc.at[i, "n8"]
        n7 = dfsrc.at[i, "n7"]
        n6 = dfsrc.at[i, "n6"]
        n5 = dfsrc.at[i, "n5"]
        n4 = dfsrc.at[i, "n4"]
        n3 = dfsrc.at[i, "n3"]
        n2 = dfsrc.at[i, "n2"]
        n1 = dfsrc.at[i, "n1"]
        p0 = dfsrc.at[i, "p0"]
        p1 = dfsrc.at[i, "p1"]
        p2 = dfsrc.at[i, "p2"]
        p3 = dfsrc.at[i, "p3"]
        p4 = dfsrc.at[i, "p4"]
        p5 = dfsrc.at[i, "p5"]
        p6 = dfsrc.at[i, "p6"]
        p7 = dfsrc.at[i, "p7"]
        p8 = dfsrc.at[i, "p8"]
        p9 = dfsrc.at[i, "p9"]
        p10 = dfsrc.at[i, "p10"]
        p11 = dfsrc.at[i, "p11"]
        p12 = dfsrc.at[i, "p12"]
        p13 = dfsrc.at[i, "p13"]

        outHigh = 0
        outHighAO = 0
        outLow = 0
        outLowAO = 0

        # Ways to find a peak that might be enhanced later with another methods using sequence recognition in machine learning
        if (
            cur > p1
            and cur > p2
            and cur > p3
            and cur > p4
            and cur > p5
            and cur > p6
            and cur > p7
            and cur > p8
            and cur > p8
            and cur > p9
            and cur > p10
            and cur > p11
            and cur > p12
            and cur > p13
        ) and (
            cur > n1
            and cur > n2
            and cur > n3
            and cur > n4
            and cur > n5
            and cur > n6
            and cur > n7
            and cur > n8
            and cur > n9
            and cur > n10
            and cur > n11
            and cur > n12
            and cur > n13
        ):
            if cur > 0 and cur > df_filterOutaof_flagThreshold_ABOVE:
                baraof_flag = 21
                if not quiet:
                    print("We have an up peak at:" + str(dt))
                countUpPeak = countUpPeak + 1
            else:
                if cur > 0 and cur > df_standardDeviation and baraof_flag != 21:
                    baraof_flag = 13
                    if not quiet:
                        print("We have an up peak at:" + str(dt))
                    countUpPeak = countUpPeak + 1
        else:
            if (
                cur < p1
                and cur < p2
                and cur < p3
                and cur < p4
                and cur < p5
                and cur < p6
                and cur < p7
                and cur < p8
                and cur < p8
                and cur < p9
                and cur < p10
                and cur < p11
                and cur < p12
                and cur < p13
            ) and (
                cur < n1
                and cur < n2
                and cur < n3
                and cur < n4
                and cur < n5
                and cur < n6
                and cur < n7
                and cur < n8
                and cur < n9
                and cur < n10
                and cur < n11
                and cur < n12
                and cur < n13
            ):
                if cur < 0 and cur < df_filterOutaof_flagThreshold_BELLOW:
                    baraof_flag = -21
                    if not quiet:
                        print("We have an down peak at: " + str(dt))
                    countDownPeak = countDownPeak + 1
                else:
                    if (
                        cur < 0
                        and cur < df_standardDeviation * -1
                        and baraof_flag != -21
                    ):
                        baraof_flag = -13
                        if not quiet:
                            print("We have an down peak at: " + str(dt))
                        countDownPeak = countDownPeak + 1
        if baraof_flag != 0:
            if cur > 0:
                outHigh = curHigh
                outHighAO = cur
                outLowAO = 0
                outLow = 0  # We are Above ZL so we dont want to use Low
            else:
                outHigh = 0  # We are Bellow ZL so we dont want to use High
                outLow = curLow
                outLowAO = cur
                outHighAO = 0

        # aof_flag is the Fractal Peak Value, it would be used to find twin peak signals and learn

        # @STCIssue I question the use of some of these columns, they might be temporary

        dfsrc.at[i, indicator_ao_fractalPeakOfMomentum_column_name] = baraof_flag
        dfsrc.at[i, indicator_ao_fractalPeakValue_column_name] = cur  # current AO Value
        dfsrc.at[i, "aofhighao"] = outHighAO
        dfsrc.at[i, "aoflowao"] = outLowAO
        dfsrc.at[i, "aofhigh"] = outHigh
        dfsrc.at[i, "aoflow"] = outLow
    l_df = len(dfsrc)
    if not quiet:
        print(
            "Total Peak - Up:"
            + str(countUpPeak)
            + ", Dn: "
            + str(countDownPeak)
            + " on total: "
            + str(l_df)
        )
    dfsrc = __cleanse_ao_peak_v1_secondary_columns(dfsrc, True)
    return dfsrc


# @title Add CDS signals

from jgtutils.jgtconstants import MFI,MFI_SQUAT,MFI_GREEN,MFI_FADE,MFI_FAKE,MFI_SIGNAL,MFI_VAL,MFI_SQUAT_STR,MFI_FAKE_STR,MFI_FADE_STR,MFI_GREEN_STR,MFI_SQUAT_ID,MFI_FAKE_ID,MFI_FADE_ID,MFI_GREEN_ID

def _cds_add_mfi_squat_n_signals_column_logics_v1(dfsrc, quiet=False):
    """
    Adds MFI (Market Facilitation Index (reference: B. Williams)) squat column logics to the given DataFrame.

    Parameters:
    - dfsrc (DataFrame): The input DataFrame to which the MFI squat column logics will be added.
    - quiet (bool): If True, suppresses the output messages. Default is False.

    Returns:
    - dfsrc (DataFrame): The updated DataFrame with the MFI squat column logics added.
    """

    dfsrc[MFI_SQUAT] = (
        (dfsrc.Volume > 0)
        & (dfsrc.Volume > dfsrc.Volume.shift())
        & (dfsrc.mfi < dfsrc.mfi.shift())
    ).astype(int)
    dfsrc[MFI_GREEN] = (
        (dfsrc.Volume > 0)
        & (dfsrc.Volume > dfsrc.Volume.shift())
        & (dfsrc.mfi > dfsrc.mfi.shift())
    ).astype(int)
    dfsrc[MFI_FADE] = (
        (dfsrc.Volume > 0)
        & (dfsrc.Volume < dfsrc.Volume.shift())
        & (dfsrc.mfi < dfsrc.mfi.shift())
    ).astype(int)
    dfsrc[MFI_FAKE] = (
        (dfsrc.Volume > 0)
        & (dfsrc.Volume < dfsrc.Volume.shift())
        & (dfsrc.mfi > dfsrc.mfi.shift())
        ).astype(int)
    dfsrc[MFI_SIGNAL] = dfsrc.apply(
         lambda row: MFI_SQUAT_ID if row[MFI_SQUAT] > 0 else MFI_GREEN_ID if row[MFI_GREEN] > 0 else MFI_FADE_ID if row[MFI_FADE]>0 else MFI_FAKE_ID if row[MFI_FAKE]>0 else 0 , axis=1
     )
    #MFI_GREEN_STR = "++"
    #MFI_FADE_STR = "--"
    #MFI_FAKE_STR = "-+"
    #MFI_SQUAT_STR = "+-"
    dfsrc[MFI_VAL] = dfsrc.apply(
         lambda row: MFI_GREEN_STR if row[MFI_SIGNAL] == MFI_GREEN_ID else MFI_FADE_STR if row[MFI_SIGNAL] == MFI_FADE_ID else MFI_FAKE_STR if row[MFI_SIGNAL] == MFI_FAKE_ID else MFI_SQUAT_STR if row[MFI_SIGNAL] == MFI_SQUAT_ID else "0" , axis=1)
    return dfsrc



def cds_add_signals_to_indicators(
    dfires,
    _aopeak_range=28,
    quiet=False,
    cc: JGTChartConfig = None,
    rq: JGTIDSRequest = None,
):
    if rq is None:
        rq = JGTIDSRequest()
    if cc is None:
        cc = JGTChartConfig()

    if not rq.disable_ao_peaks_v1:
        dfires = _add_ao_fractal_peak_v1(
            dfires,
            AO,
            "p",
            "n",
            _aopeak_range,
            quiet=quiet,
            cc=cc,
        )
        dfires = __cleanse_ao_peak_v1_secondary_columns(dfires, quiet=True)
    return dfires


def tocds(
    dfsrc,
    quiet=True,
    cc: JGTChartConfig = None,
    rq: JGTIDSRequest = None,
    columns_to_remove=None,
    use_v2_jgtapyhelper=True,
    add_mfi_signals_proto=True,
):
    if rq is None:
        rq = JGTIDSRequest()
    if cc is None:
        cc = JGTChartConfig()
    if use_v2_jgtapyhelper:
        import jgtapyhelper as tah
        #print("RQ:", rq.timeframe)
        dfires =tah.toids(dfsrc, cc=cc, rq=rq, quiet=quiet, columns_to_remove=columns_to_remove)
    else:
        print("DEPRECATION NOTICE: tocds() is deprecated and will be removed in a future release. Please use toids() from jgtapyhelper instead.")
        dfires = ids_add_indicators(dfsrc, quiet=quiet, cc=cc, rq=rq)
    
    dfires = _ids_add_fdb_column_logics_v2(dfires, quiet=quiet)                 #@STCIssue SignalBusiness Code
    dfires = cds_add_signals_to_indicators(dfires, quiet=quiet, cc=cc, rq=rq)   #@STCIssue SignalBusiness Code
    dfires = jgti_add_zlc_plus_other_AO_signal(dfires, quiet=quiet, rq=rq)      #@STCIssue SignalBusiness Code
    if rq.keep_bid_ask==False:
        dfires = _pds_cleanse_original_columns(dfires, quiet=True)  
    dfires = __cleanse_ao_peak_v1_secondary_columns(dfires, quiet=True)
    dfires = __format_boolean_columns_to_int(dfires, quiet=True)
    dfires = add_ao_price_peaks_v2(dfires, quiet=True, rq=rq)   
    
    if add_mfi_signals_proto:
        dfires= _cds_add_mfi_squat_n_signals_column_logics_v1(dfires, quiet=quiet)
    
    
    # Remove the specified columns
    if columns_to_remove is not None:
        dfires.drop(columns=columns_to_remove, errors="ignore", inplace=True)
    return dfires


def __format_boolean_columns_to_int(dfsrc, quiet=True):
    for col in dfsrc.columns:
        if dfsrc[col].dtype == bool:
            dfsrc[col] = dfsrc[col].astype(int)
    return dfsrc




# @title ZLC Buy and Sell v2 2210161707


def jgti_add_zlc_plus_other_AO_signal(
    dfsrc,
    dropsecondaries=True,
    quiet=True,
    rq: JGTIDSRequest = None,
):
    if rq is None:
        rq = JGTIDSRequest()

    dfsrc = _jgtpd_col_add_range_shifting(dfsrc, AO, "pao", 10)
    dfsrc = _jgtpd_col_add_range_shifting(dfsrc, AC, "pac", 4)
    if not quiet:
        print("----added shofted range AO")

    # AO Above Zero
    dfsrc[AOAZ] = (dfsrc[AO] > 0).astype(int)
    # AO Bellow Zero
    dfsrc[AOBZ] = (dfsrc[AO] < 0).astype(int)

    c = 0
    xc = len(dfsrc)
    for i, row in dfsrc.iterrows():
        c = c + 1
        cao = dfsrc.at[i, AO]  # Current AO
        cac = dfsrc.at[i, AC]  # Current AC
        pac1 = dfsrc.at[i, "pac1"]  # Past AC 1
        pac2 = dfsrc.at[i, "pac2"]  # Past AC 2
        pac3 = dfsrc.at[i, "pac3"]  # Past AC 3
        cacgreen = False
        pac1green = False
        pac2green = False
        if cac > pac1:
            cacgreen = True
        if pac1 > pac2:
            pac1green = True
        if pac2 > pac3:
            pac2green = True

        pao1 = dfsrc.at[i, "pao1"]  # Past AO 1
        pao2 = dfsrc.at[i, "pao2"]  # Past AO 2
        pao3 = dfsrc.at[i, "pao3"]  # Past AO 3
        caogreen = False
        pao1green = False
        pao2green = False
        if cao > pao1:
            caogreen = True
        if pao1 > pao2:
            pao1green = True
        if pao2 > pao3:
            pao2green = True

        # For simplicity
        caored = not caogreen
        pao1red = not pao1green
        pao2red = not pao2green

        cacred = not cacgreen
        pac1red = not pac1green
        pac2red = not pac2green

        aoaz = dfsrc.at[i, AOAZ]
        aobz = dfsrc.at[i, AOBZ]

        # ZLC
        isZLCBuy = 0
        isZLCSell = 0
        zlcCode = 0
        if pao1 > 0 and aobz == 1:
            zlcCode = -1
            isZLCSell = 1
        if pao1 < 0 and aoaz == 1:
            zlcCode = 1
            isZLCBuy = 1

        dfsrc.at[i, ZLC] = zlcCode
        dfsrc.at[i, ZLCB] = isZLCBuy
        dfsrc.at[i, ZLCS] = isZLCSell

        # dfsrc[signal_zcol_column_name] = dfsrc[signal_zcol_column_name].astype(object)
        # Coloring AO
        # dfsrc['aocolor'] = dfsrc['aocolor'].astype(object)

        if rq.include_ao_color:
            if caogreen:
                dfsrc.at[i, "aocolor"] = "green"
            else:
                dfsrc.at[i, "aocolor"] = "red"

        # Coloring AC
        # dfsrc['accolor'] = dfsrc['accolor'].astype(object)
        if rq.include_ac_color:
            if cacgreen:
                dfsrc.at[i, "accolor"] = "green"
            else:
                dfsrc.at[i, "accolor"] = "red"

        # --@STCIssue Zone  (Not sure, it might have to be ABove or Bellow)

        zoneColor = nonTradingZoneColor  # default Zone Color

        redZone = 0
        if cacred and caored and pac1red and pao1red:
            redZone = 1
            zoneColor = sellingZoneColor

        greenZone = 0
        if cacgreen and caogreen and pac1green and pao1green:
            greenZone = 1
            zoneColor = buyingZoneColor

        dfsrc.at[i, signal_zcol_column_name] = zoneColor

        # Sell Zone Signal

        dfsrc.at[i, SZ] = redZone

        # Buy Zone Signal
        dfsrc.at[i, BZ] = greenZone

        # AC Sell / Buy  3 AC Against AO af AC Bellow, 2 if above
        acSell = 0
        msgacSignal = "No "
        if cacred and pac1red and caogreen and pao1green:
            acSell = 1
            if (
                cac < 0 and pac2green
            ):  # We require 3 bars red on the AC When bellow zero
                acSell = 0
        acBuy = 0
        if cacgreen and pac1green and caored and pao1red:
            acBuy = 1
            if cac > 0 and pac2red:
                acBuy = 0

        # AC Sell Signal (Deceleration)

        dfsrc.at[i, ACS] = acSell

        # AC Buy Signal (Acceleration)
        dfsrc.at[i, ACB] = acBuy

        if acSell and not quiet:
            print("AC Sell Signal with AC Bellow Zero Line " + str(i))
        if acBuy and not quiet:
            print("AC Buy Signal with AC ABove Zero Line " + str(i))

        # Saucer Strategy
        # More on Saucer Strategy : http://simp.ly/p/2K1HBr
        saucerSell = 0
        if cao < 0 and caored and pao1green and pao2green:
            saucerSell = 1

        saucerBuy = 0
        if cao > 0 and caogreen and pao1red and pao2red:
            saucerBuy = 1

        dfsrc.at[i, SS] = saucerSell
        dfsrc.at[i, SB] = saucerBuy

        # What Happens on the Next PLUS 35 Periods ??
        if c < xc - 35:
            cPrice = row["Close"]
    return dfsrc





"""
This module, JGTCDSSvc.py, is part of a larger application designed to interact with the Chaos Data Service (CDS). It includes functionality to create data requests for financial market data based on user-specified parameters. The module handles the initialization and configuration of these data requests, leveraging the Chaos Data Service's capabilities to fetch and process financial data.

Imports:
- warnings: Used to ignore FutureWarning messages that might arise.
- sys, os: For path manipulations and system-level operations.
- pandas (pd): For handling data structures and operations on them.
- json: For parsing JSON data.

The module also imports specific functionalities from other parts of the application:
- JGTCDS: A module that likely contains core functionalities related to the Chaos Data Service.
- JGTCDSRequest: A class definition for creating request objects to interact with the CDS.
- jgtutils.jgtpov: Utilities for processing or obtaining higher timeframes from given data.
- jgtutils.jgtos: Operating system utilities, such as path getters and creators.
- jgtutils.jgtconstants (c): Constants used throughout the application, possibly including default values or configuration settings.

Functions:
- createCDSRequestFromArgs(args, instrument, timeframe): This function is the primary interface for creating a CDS request object. It takes command line arguments, a specified financial instrument, and a timeframe to create a JGTCDSRequest object. This object is then configured with various options such as whether to include bid/ask data, the number of quotes to retrieve, and specific analytical flags like the inclusion of the Gator Oscillator or the Money Flow Index. The function is well-documented with a detailed docstring explaining its parameters, functionality, and return value.

Overall, this module serves as a bridge between the user's command-line inputs and the Chaos Data Service, enabling the dynamic creation and configuration of data requests based on user preferences and requirements.
"""

import warnings

# Ignore FutureWarning
warnings.simplefilter(action="ignore", category=FutureWarning)

import sys
import os
import pandas as pd
import json


sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

import JGTCDS as cds
from JGTCDSRequest import JGTCDSRequest
from jgtutils.jgtpov import get_higher_tf_by_level,get_higher_tf_array
from jgtutils.jgtos import get_data_path,mk_fullpath
from jgtutils import jgtconstants as c

def createCDSRequestFromArgs(args, instrument, timeframe):
    """
    Creates a Chaos Data Service (CDS) request object from command line arguments.

    This function initializes a JGTCDSRequest object with parameters specified through command line arguments for requesting data from the Chaos Data Service. The request object is configured with various options such as the financial instrument, timeframe, and additional flags for data processing.

    Parameters:
    - args (Namespace): The parsed command line arguments containing the options for the CDS request.
    - instrument (str): The financial instrument for which the data is requested (e.g., "EUR/USD", "Gold").
    - timeframe (str): The timeframe for the data request (e.g., "1m", "1h").

    The function configures the JGTCDSRequest object with the following options:
    - instrument: The financial instrument for the request.
    - keep_bid_ask: A boolean indicating whether to keep bid/ask data. It is set to False if args.rmbidask is True, indicating the removal of bid/ask data.
    - timeframe: The timeframe for the data request.
    - quotescount: The number of quotes to retrieve. Defaults to 300 if not specified.
    - viewpath: The path for the view. Defaults to False if not specified.
    - use_fresh: A boolean indicating whether to use fresh data. Defaults to False if not specified or if args.notfresh is True, indicating stale data may be used.
    - use_full: A boolean indicating whether to use full data. Defaults to False if not specified.
    - gator_oscillator_flag: A boolean indicating whether to include the Gator Oscillator in the data. Defaults to False if not specified.
    - mfi_flag: A boolean indicating whether to include the Money Flow Index in the data. Set to False if args.no_mfi_flag is True.
    - balligator_flag: A boolean indicating whether to include the Bollinger Alligator in the data. Defaults to False if not specified.
    - talligator_flag: A boolean indicating whether to include the T Alligator in the data. Defaults to False if not specified.
    - balligator_period_jaws: The period for the jaws of the Bollinger Alligator. This option is currently not linked to other balance lines, which may limit its usefulness.
    - largest_fractal_period: The largest period for fractals to be considered in the data.
    - verbose_level: The verbosity level for logging purposes.

    Returns:
    - JGTCDSRequest: The initialized CDS request object.
    """
    rq = JGTCDSRequest()
    rq.instrument = instrument
    #rq.keep_bid_ask = args.keepbidask
    if args.rmbidask:
        rq.keep_bid_ask = False
    rq.timeframe = timeframe
    rq.quotescount = args.quotescount if args.quotescount else 300
    rq.viewpath=args.viewpath if args.viewpath else False
    rq.use_fresh = args.fresh if args.fresh else False
    if args.notfresh:
        rq.use_fresh = False
    rq.use_full = args.full if args.full else False
    rq.gator_oscillator_flag = (
        args.gator_oscillator_flag if args.gator_oscillator_flag else False
    )
    #rq.mfi_flag = args.mfi_flag if args.mfi_flag else False
    if args.no_mfi_flag:
        rq.mfi_flag = False
    rq.balligator_flag = args.balligator_flag if args.balligator_flag else False
    rq.talligator_flag = args.talligator_flag if args.talligator_flag else False
    rq.balligator_period_jaws = args.balligator_period_jaws #@STCIssue That is useless not to have others balance lines
    rq.largest_fractal_period = args.largest_fractal_period
    rq.verbose_level = args.verbose
    return rq


def set_rq_defaults(rq):
    rq.keep_bid_ask = True
    if not rq.use_full and rq.quotescount == -1:
      rq.quotescount = 300
    rq.viewpath = False
    rq.gator_oscillator_flag = False
    rq.mfi_flag = True
    rq.balligator_flag = True
    rq.talligator_flag = True
    
    if rq.timeframe=="M1":
      rq.talligator_flag = False
      
    rq.balligator_fix_quotescount()
    rq.talligator_fix_quotescount()
    return rq

    
    
def new_rq_default(instrument,timeframe,use_fresh=True,use_full=False,quotescount=-1):
    rq = JGTCDSRequest()
    rq.instrument = instrument
    rq.timeframe = timeframe
    rq.quotescount = quotescount
    rq.use_full = use_full
    rq.use_fresh = use_fresh
    rq=set_rq_defaults(rq)
    # if rq.talligator_flag:
    #   print("Talligator flag is set")
    return rq

def get(instrument,timeframe,use_full=False,use_fresh=True,quotescount=-1,quiet=True,dont_write_cds=False):
    """
    Generate the CDS to file
    """
    # Parse command line arguments
    # args = parse_args()
    # instrument = args.instrument
    # timeframe = args.timeframe

    # Create CDS request
    rq = new_rq_default(instrument,timeframe,use_fresh=use_fresh,use_full=use_full,quotescount=quotescount)

    # Create CDS
    cdf= cds.create2(rq,quiet=quiet)
    if not dont_write_cds:
      cds.writeCDS(instrument, timeframe, use_full, cdf)
      #cds.writeCDS(cdf=cdf,instrument=instrument,timeframe=timeframe,use_full=use_full,quotescount=quotescount,quiet=quiet)
    return cdf

def read(instrument,timeframe,use_full=False,quotescount=-1,quiet=True):
  try:
    cdf= cds.readCDSFile(instrument,timeframe,use_full=use_full,quote_count=quotescount,quiet=quiet)
  except:
    cdf=None
  if cdf is None:
    print("CDS Could not read, so we create it....",timeframe)
    cdf=get(instrument,timeframe,use_full=use_full,quotescount=quotescount,quiet=quiet)
  #print("CDSSvc Read: cdf len: ", len(cdf))
  return cdf

import concurrent.futures

# Move the get_cdf function to the top level
def get_cdf(i, t, tf, use_full, use_fresh, quiet, quotescount,force_read=False):
  if use_fresh:
    force_read=False
  _f=":f" if use_full else ""
  if force_read:
    print("   CDSSvc Read: ",i," ", tf, "/", t, _f,"\n")
    return tf, read(i, tf, use_full=use_full, quotescount=quotescount, quiet=quiet)
  else:
    print("CDSSvc Create: ",i," ", tf, "/", t, _f,"\n")
    return tf, get(i, tf, use_full=use_full, use_fresh=use_fresh, quiet=quiet, quotescount=quotescount)

def get_higher_cdf_datasets(i, t, use_full=False, use_fresh=True, quiet=True, quotescount=-1,force_read=False):
  if not use_fresh and not force_read:
    force_read=True # We force read if not use_fresh
    
  tf_array = get_higher_tf_array(t)
  if not quiet:
    print("Higher TF Array: ", tf_array)
  res = {}

  # Use a ProcessPoolExecutor to run the function in parallel for each tf in tf_array
  with concurrent.futures.ProcessPoolExecutor() as executor:
    futures = {executor.submit(get_cdf, i, t, tf, use_full, use_fresh, quiet, quotescount,force_read) for tf in tf_array}

    for future in concurrent.futures.as_completed(futures):
      tf, cdf = future.result()
      res[tf] = cdf

  return res

def get_higher_cdf_datasets__ThreadPoolExecutor(i, t, use_full=False, use_fresh=True, quiet=True, quotescount=300):
  tf_array = get_higher_tf_array(t)
  if not quiet:
    print("Higher TF Array: ", tf_array)
  res = {}

  # Define a function to be run in parallel
  def get_cdf(tf):
    print("CDSSvc Get: ", tf, " of : ", t," for: ",i)
    return tf, get(i, tf, use_full=use_full, use_fresh=use_fresh, quiet=quiet, quotescount=quotescount)

  # Use a ThreadPoolExecutor to run the function in parallel for each tf in tf_array
  with concurrent.futures.ThreadPoolExecutor(max_workers=10) as executor:
    futures = {executor.submit(get_cdf, tf) for tf in tf_array}

    for future in concurrent.futures.as_completed(futures):
      tf, cdf = future.result()
      res[tf] = cdf

  return res

def get_higher_cdf_datasets_no_concurrence(i,t,use_full=False,use_fresh=True,quiet=True,quotescount=300):
  tf_array=get_higher_tf_array(t)
  if not quiet:
    print("Higher TF Array: ",tf_array)
  res={}
  for tf in tf_array:
    #if not quiet:
    print("CDSSvc Get: ",tf," of : ",t)
    cdf=get(i,tf,use_full=use_full,use_fresh=use_fresh,quiet=quiet,quotescount=quotescount)
    res[tf]=cdf
    #res.append(cdf)
  return res

def get_higher_cdf(i,t,_level=1,default_timeframes = "M1,W1,D1,H4,H1,m15,m5",quiet=True):
  if default_timeframes=="T":# try read the os var "T"
    default_timeframes=os.getenv("T","M1,W1,D1,H4,H1,m15,m5")
    if not quiet:
       print("default timeframes read from env: ",default_timeframes)
  
  if _level==0:
    return get(i,t),t # Current
  else:
    htf=get_higher_tf_by_level(t,_level,default_timeframes)
    if not quiet:
       print("Higher TF: ",htf," of : ",t)
    if htf==None:
      return None,None # No higher TF
    

    hdf=get(i,htf)
    
    return hdf,htf


def zone_get_value_from_cdf(_cdf):
  return _cdf[c.ZCOL].tail(1).values[0]

def _get_zone_index_from_cdf(_cdf):
  return _cdf[c.ZCOL].tail(1).index[0]

def zone_update(i,t,quiet=True,to_json=False): # We expect it to be saved in the zone folder
  cdf=get(i,t)
  return zone_update_from_cdf(i,t,cdf,quiet=quiet,to_json=to_json)

def zone_update_from_cdf(i,t,cdf,quiet=True,to_json=False): # We expect it to be saved in the zone folder  
  zone=zone_get_value_from_cdf(cdf)
  if not quiet:
    print("Zone: ",zone)
  
  zone_data = zone__from_cdf_to_zonedata(cdf)
  # Save the zone
  fpath=zone_save_data(i,t,zone_data)
  if to_json:
    r= {}
    r["path"]=fpath
    
    # try:
    #    r["data"]=zone_data
    # except:
    #   r["data"]=zone_data

    return json.dumps(r)
  return fpath,zone_data

#@STGoal Current and Higher timeframes zone are updated.
def zone_update_higher(i,t,level=1,quiet=True,to_json=False): # We expect it to be saved in the zone folder
  res={}
  current_path,cur_zone_data = zone_update(i,t,quiet=quiet)
  cur={}
  cur["data"]=cur_zone_data
  cur["path"]=current_path
  res[t]=cur
  
  if level>0:
    # #Update the first level zone
    # if not quiet:print("Update the first zone level:",t)

    # htf1=get_higher_tf_by_level(t,level)
    # print("Higher TF: ",htf1," of : ",t)
    # if htf1==None:
    #   return res
    # tf1_path=zone_update(i,htf1)
    # res.append(tf1_path)
    #iterate thru level 
    for ii in range(1,level+1):
      htf=get_higher_tf_by_level(t,ii)
      if htf is None:
        continue
      if not quiet:print(f"Update the {ii} zone level ({t}/{htf})")
      tf_path,_zd=zone_update(i,htf,quiet=quiet,to_json=to_json)
      hz={}
      hz["data"]=_zd
      hz["path"]=tf_path
      res[htf]=hz
  if to_json:
    return json.dumps(res)
  return res
  
# Get the zone from data store
def zone_read(i,t,add_tf_key=True,to_json=False,add_instrument_key=False):
  data_path_zone = get_data_path("zone")
  fpath = mk_fullpath(i, t, "csv", data_path_zone)
  #if not exist fpath
  if not os.path.exists(fpath):
    print("No zone data found for ",i,t)
    print("Creating zone data for ",i,t)
    zone_update(i,t)
  zone_data=pd.read_csv(fpath)
  if add_tf_key:
    zone_data["t"]=t
  if add_instrument_key:
    zone_data["i"]=i
  if to_json:
    return df_to_json(zone_data)
  return zone_data



def df_to_json(df,row_id=0,orient='records'):
  json_str = df.to_json(orient=orient)
  json_obj = json.loads(json_str)
  return json_obj[row_id] if json_obj else None

def read_zone_to_json(i, t,add_instrument_key=False,add_tf_key=False):
  return zone_read(i, t, to_json=True,add_instrument_key=add_instrument_key,add_tf_key=add_tf_key)



#read zone and higher timeframe level
def zone_read_up(i,t,level=1,quiet=False,to_json=False,add_instrument_key=False,add_tf_key=False):
  res={}
  # cur_zdata=read_zone(i,t).to_dict(orient="records")
  zone_data = zone_read(i,t,add_tf_key=add_tf_key,add_instrument_key=add_instrument_key)
  cur_zdata=df_to_json(zone_data)
  res[t]=cur_zdata
  #res.append(current)
  if level>0:
    # #Update the first level zone
    # if not quiet:print("Read the first zone level:",t)

    # tf1=get_higher_tf_by_level(t,level)
    # print("Higher TF: ",tf1," of : ",t)
    # if tf1==None:
    #   return res
    # h1_zdata=read_zone(i,tf1)
    # res[tf1]=h1_zdata
    #iterate thru level 
    for ii in range(1,level+1):
      tf_ii=get_higher_tf_by_level(t,ii)
      if not quiet:print(f"Read the {ii} zone level ({t}/{tf_ii})")
      # hz_data=read_zone(i,tf_ii).to_dict(orient="records")
      hz_data=df_to_json(zone_read(i,tf_ii,add_tf_key=False))
      res[tf_ii]=hz_data
  if to_json:
    return json.dumps(res)
  return res


def zone__from_cdf_to_zonedata(cdf,add_zone_int=True): # expect to keep one column,last row
    zone_data=cdf.tail(1)
    #print(zone_data)
    #keep c.ZCOL in zone_data
    zone_data=zone_data[[c.ZCOL]]

    if add_zone_int:
      zone_data[c.ZONE_INT]=c.ZONE_BUY_ID if zone_data[c.ZCOL].values[0] == c.ZONE_BUY_STR else c.ZONE_SELL_ID if zone_data[c.ZCOL].values[0] == c.ZONE_SELL_STR else c.ZONE_NEUTRAL_ID
    
    return zone_data


def zone_save_data(instrument, timeframe, zone_data,save_json=True):
    data_path_zone = get_data_path("zone")
    fpath = mk_fullpath(instrument, timeframe, "csv", data_path_zone)
    # print(fpath)
    zone_data.to_csv(fpath, index=True)
    fpath_json = mk_fullpath(instrument, timeframe, "json", data_path_zone)
    
    if save_json:
      zone_data_json=df_to_json(zone_data)
      with open(fpath_json, 'w') as outfile:
          json.dump(zone_data_json, outfile)

    return fpath




#!/usr/bin/env Python
"""
jgtpy
"""

from setuptools import find_packages, setup

#from jgtpy import __version__ as version
def read_version():
    with open("jgtpy/JGTCore.py") as f:
        for line in f:
            if line.startswith("__version__"):
                return line.strip().split()[-1][1:-1]

version = read_version()


setup(
    name="jgtpy",
    version=version,
    description="JGTrading Data maker' Dataframes",
    long_description=open("README.rst").read(),
    author="GUillaume Isabelle",
    author_email="jgi@jgwill.com",
    url="https://github.com/jgwill/jgtpy",
    packages=find_packages(
        include=["jgtpy", "test-*.py"], exclude=["test*log", "*test*csv", "*test*png"]
    ),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Financial and Insurance Industry",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python :: 3.7.16",
    ],
)
